//
// JavaVM - interpret instrukc
//
# include "javavm.h"
# include <string.h>
# include <stdio.h>
# include <stdarg.h>
# include <stdlib.h>

void CHECK(bool cond, char* msg, ...)
{
   if( !cond ) {
      va_list ap;
      va_start(ap, msg);
      fprintf(stderr, "VM Error: ");
      vfprintf(stderr, msg, ap);
      fprintf(stderr, "\n");
      abort();
   }
} // CHECK

class CJavaArray {
public:
   char       elemType;
   int        numElems;
   long*      data;

   CJavaArray(char t, int n);
   long get(long index);
   void put(long index, long value);
}; // CJavaArray

CJavaArray::CJavaArray(char t, int n)
{
   elemType = t;
   numElems = n;
   if( n > 0 ) {
      data = new long[ n ];
      memset(data, 0, n * sizeof(long));
   } else
      data = NULL;
} // CJavaArray::CJavaArray

long CJavaArray::get(long index)
{
   if( index < 0 || index >= numElems ) {
      CHECK(false, "Array index out of bounds 0..%u", numElems-1);
      return 0;
   }
   return data[index];
} // CJavaArray::get

void CJavaArray::put(long index, long value)
{
   if( index < 0 || index >= numElems ) {
      CHECK(false, "Array index out of bounds 0..%u", numElems-1);
      return;
   }
   data[index] = value;
} // CJavaArray::put

//
// JavaState
// =========

JavaState::JavaState(JavaMethod* method)
{
   code.open(method);
   code.getLimits(maxStack, maxLocals);
   locals = new JavaInt[maxStack+maxLocals];
// memset(locals, 0, (maxStack+maxLocals)*sizeof(JavaInt));
   stack = locals + maxLocals;
   top = 0;
   currentPC = 0;
}

JavaState::~JavaState()
{
   code.close();
   delete locals;
} // JavaState::~JavaState

void JavaState::save()
{
   currentPC = code.getPC();
} // JavaState::save

void JavaState::restore()
{
   code.setPC(currentPC);
} // JavaState::restore

void JavaState::push(JavaInt val)
{
   CHECK( top < maxStack, "push na plny zasobnik" );
   stack[top++] = val;
} // JavaState::push

JavaInt JavaState::pop()
{
   CHECK( top >= 0, "pop na prazdny zasobnik" );
   return stack[--top];
} // JavaState::pop

//
// JavaVM
// ======

JavaVM::JavaVM()
{
} // JavaVM::JavaVM

void JavaVM::stdIoCall(JavaCString methodName, JavaCString methodDescr)
{
   if( strcmp(methodName, "printString") == 0 ) {
      JavaCString str = (JavaCString)state->pop();
      fprintf(outFile, "%s", str);
   } else if( strcmp(methodName, "printInt") == 0 ) {
      fprintf(outFile, "%d\n", state->pop());
   } else if( strcmp(methodName, "readInt") == 0 ) {
      JavaInt val = 0;
      fprintf(outFile, "? ");
      fscanf(inpFile, "%d", &val);
      if( inpFile != stdin )
         fprintf(outFile, "%d\n", val);
      state->push(val);
   } else if( strcmp(methodName, "readChar") == 0 ) {
      JavaInt val = getchar();
      state->push(val == EOF ? 0 : val);
   } else if( strcmp(methodName, "printChar") == 0 ) {
      JavaInt val = state->pop();
      putchar(val);
   } else if( strcmp(methodName, "printLn") == 0 ) {
      fprintf(outFile, "\n");
   } else
      CHECK(false, "Neznama metoda 'io.%s%s'", methodName, methodDescr);
} // stdIoCall

JavaUShort skipDescr(JavaCString& descr)
{
   JavaUShort size = 0;
   JavaCString savedDescr = descr;
   switch( *descr ) {
      case JavaClass::TYPE_INT: case JavaClass::TYPE_FLOAT: 
         size = 1;
         descr++;
         break;
      case JavaClass::TYPE_LONG: case JavaClass::TYPE_DOUBLE:
         size = 2;
         descr++;
         break;
      case JavaClass::TYPE_ARRAY:
         size = 1;
         descr++;
         skipDescr(descr);
         break;
      case JavaClass::TYPE_CLASS:
         size = 1;
         descr++;
         while(*descr != ';') {
            CHECK( *descr != '\0', "Neplatny deskriptor '%s'", savedDescr );
            descr++;
         }
         descr++;
         break;
      case JavaClass::TYPE_VOID:
         size = 0;
         descr++;
         break;
      default:
         CHECK( false, "Neplatny deskriptor '%s'", savedDescr );
         break;
   }
   return size;
} // skipDescr

void JavaVM::procFieldOps(JavaInstr op, JavaUShort indx)
{
   const JavaConst* cp = cls->getConst(indx);
   const char* className;
   const char* fieldName;
   const char* fieldDescr;
   cp->getRef(className, fieldName, fieldDescr);
   if( strcmp(className, cls->getClassName()) == 0 ) {
      JavaField* field = cls->getField(fieldName, fieldDescr);
      CHECK(field != NULL, "Neznama promenna '%s.%s' typu '%s'", className, fieldName, fieldDescr);
      if( op == GETSTATIC )
         state->push(field->value().i);
      else
         field->value().i = state->pop();
   } else {
      CHECK(false, "Odkaz na promennou cizi tridy '%s.%s'", className, fieldName);
   }
} // procFieldOps

void JavaVM::procInvokeOps(JavaInstr op, JavaUShort indx)
{
   const JavaConst* cp = cls->getConst(indx);
   const char* className;
   const char* methodName;
   const char* methodDescr;
   cp->getRef(className, methodName, methodDescr);

   if( strcmp(className, cls->getClassName()) == 0 ) {
      
      // vyhledme sprvnou metodu
      JavaMethod* method = cls->getMethod(methodName, methodDescr);
      CHECK(method != NULL, "Neznama metoda '%s.%s%s'", className, methodName, methodDescr);

      // spoteme velikost argument
      JavaUShort argc = 0;
      JavaCString descr = methodDescr;
      CHECK( *descr == '(', "Neplatny deskriptor '%s'", methodDescr );
      descr++;
      while( *descr != ')' ) {
         argc += skipDescr(descr);
      }
      CHECK( *descr == ')', "Neplatny deskriptor '%s'", methodDescr );
      descr++;

      // nastavme zsobnk na argumenty
      CHECK( argc <= state->top, "Malo argumentu na zasobniku pri volani '%s%s'", methodName, methodDescr);
      state->top -= argc;
      JavaInt* argv = &state->stack[state->top];

      // ulome aktuln stav vpotu a pipravme nov aktivan zznam
      pushState(method);
   
      // pedme argumenty
      CHECK( argc <= state->maxLocals, "Malo mista pro argumenty pri volani '%s%s'", methodName, methodDescr);
      memcpy(state->locals, argv, argc * sizeof(JavaInt));
   } 
   else if( strcmp(className, "io") == 0 )
      stdIoCall(methodName, methodDescr);
   else {
      CHECK(false, "Odkaz na metodu cizi tridy '%s.%s'", className, methodName);
   }
} // procInvokeOps

void JavaVM::procNewArray(JavaUByte indx)
{
   char t = JavaClass::TYPE_INT;
   switch( indx ) {
   case  4: t = JavaClass::TYPE_BOOLEAN; break;
   case  5: t = JavaClass::TYPE_CHAR;    break;
   case 10: t = JavaClass::TYPE_INT;     break;
   default:
      CHECK(false, "NewArray pro nepodporovany typ %u", indx);
      break;
   }
   long count = state->pop();
   CJavaArray* ap = new CJavaArray(t, count);
   state->push((long)ap);
} // JavaVM::procNewArray

void JavaVM::run(JavaClass& mainClass, FILE* iFile, FILE* oFile)
{
   cls = &mainClass;
   state = NULL;
   inpFile = ( iFile == NULL ) ? stdin : iFile;
   outFile = ( oFile == NULL ) ? stdout : oFile;

   JavaMethod* main = cls->getMethod(mainName, mainDescr);
   CHECK( main != NULL, "Nenalezena metoda '%s%s'", mainName, mainDescr );

   pushState(main);

   running = true;
   while( running ) {
      step();
   }

} // JavaVM::main

void JavaVM::procReturnOps(JavaInstr op)
{
   switch( op ) {
      case IRETURN:
         result.i = state->pop();
         break;
      case RETURN:
         break;
      default:
         CHECK(false, "procReturnOps");
         break;
   }

   popState();

   // pevezmeme vsledek
   switch( op ) {
      case IRETURN:
         state->push(result.i);
         break;
      case RETURN:
         break;
      default:
         CHECK(false, "procReturnOps");
         break;
   }

   if( state == NULL )
      running = false;
} // JavaVM::procReturnOps

void JavaVM::pushState(JavaMethod* method)
{
   if( state != NULL )
      state->save();

   JavaState* newState = new JavaState(method);
   newState->previous = state;
   state = newState;
   state->restore();
} // JavaVM::pushState

void JavaVM::popState()
{
   JavaState* oldState = state;
   state->save();
   state = state->previous;
   delete oldState;

   if( state != NULL )
      state->restore();
} // JavaVM::popState

void JavaVM::step()
{
   bool wideFlag = false;

AGAIN:
   JavaUShort basePC = state->code.getPC();
   if( basePC == state->code.getCodeSize() ) {
      fprintf(stderr, "JVM Warning: Chybi RETURN\n");
      procReturnOps(RETURN);
      return;
   }

   JavaInstr oc = (JavaInstr)(JavaUByte)state->code.fetchByte();
   switch( oc ) {
   case NOP:
      break;
   case ICONST_M1: case ICONST_0: case ICONST_1: case ICONST_2:
   case ICONST_3: case ICONST_4: case ICONST_5:
      state->push( (JavaInt)oc - ICONST_0 );
      break;
   case BIPUSH:
      state->push(state->code.fetchByte());
      break;

   case SIPUSH:
      state->push(state->code.fetchShort());
      break;

   case IADD: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      state->push(left + right);
      break;
   }

   case ISUB: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      state->push(left - right);
      break;
   }

   case IMUL: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      state->push(left * right);
      break;
   }

   case IDIV: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      CHECK( right != 0, "Deleni nulou v instrukci IDIV" );
      state->push(left / right);
      break;
   }

   case IREM: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      CHECK( right != 0, "Deleni nulou v instrukci IREM" );
      state->push(left % right);
      break;
   }

   case INEG: {
      JavaInt left = state->pop();
      state->push(-left);
      break;
   }

   case IAND: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      state->push(left & right);
      break;
   }

   case IOR: {
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      state->push(left | right);
      break;
   }

   case GOTO: {
      JavaUShort newPC = basePC + state->code.fetchShort();
      state->code.setPC(newPC);
      break;
   }

   case IF_ICMPEQ: case IF_ICMPNE: case IF_ICMPLT:
   case IF_ICMPGT: case IF_ICMPGE: case IF_ICMPLE: {
      JavaUShort newPC = basePC + state->code.fetchShort();
      JavaInt right = state->pop();
      JavaInt left = state->pop();
      bool ok;
      if( left == right )
         ok = (oc == IF_ICMPEQ || oc == IF_ICMPGE || oc == IF_ICMPLE);
      else if( left < right )
         ok = (oc == IF_ICMPNE || oc == IF_ICMPLT || oc == IF_ICMPLE);
      else
         ok = (oc == IF_ICMPNE || oc == IF_ICMPGT || oc == IF_ICMPGE);
      if( ok )
         state->code.setPC(newPC);
      break;
   }

   case IFEQ: case IFNE: case IFLT: case IFGT: case IFGE: case IFLE: {
      JavaUShort newPC = basePC + state->code.fetchShort();
      JavaInt left = state->pop();
      bool ok;
      if( left == 0 )
         ok = (oc == IFEQ || oc == IFGE || oc == IFLE);
      else if( left < 0 )
         ok = (oc == IFNE || oc == IFLT || oc == IFLE);
      else
         ok = (oc == IFNE || oc == IFGT || oc == IFGE);
      if( ok )
         state->code.setPC(newPC);
      break;
   }

   case GETSTATIC: case PUTSTATIC: {
//    case GETFIELD: case PUTFIELD: {
      JavaUShort indx = state->code.fetchShort();
      procFieldOps(oc, indx);
      break;
   }

   case INVOKESTATIC: case INVOKEVIRTUAL: case INVOKESPECIAL: {
      JavaUShort indx = state->code.fetchShort();
      procInvokeOps(oc, indx);
      break;
   }

//      case LDC_W:
   case LDC: {
      JavaUByte indx = state->code.fetchByte();
      const JavaConst* cp = cls->getConst(indx);
      switch( cp->getTag() ) {
         case JavaConst::CONSTANT_Integer:
            state->push(cp->getInteger());
            break;
         case JavaConst::CONSTANT_String:
            state->push((JavaInt)cp->getString());
            break;
         default:
            CHECK( false, "Neimplementovana varianta instrukce LDC" );
      }
      break;
   }

   case ILOAD_0: case ILOAD_1: case ILOAD_2: case ILOAD_3:
      state->push(state->locals[oc-ILOAD_0]);
      break;

   case ILOAD: {
      JavaUInt indx;
      if( wideFlag )
         indx = state->code.fetchShort();
      else
         indx = state->code.fetchByte();
      state->push(state->locals[indx]);
      break;
   }

   case ALOAD_0: case ALOAD_1: case ALOAD_2: case ALOAD_3:
      state->push(state->locals[oc-ALOAD_0]);
      break;

   case ALOAD: {
      JavaUInt indx;
      if( wideFlag )
         indx = state->code.fetchShort();
      else
         indx = state->code.fetchByte();
      state->push(state->locals[indx]);
      break;
   }

   case ISTORE_0: case ISTORE_1: case ISTORE_2: case ISTORE_3:
      state->locals[oc-ISTORE_0] = state->pop();
      break;

   case ISTORE: {
      JavaUInt indx;
      if( wideFlag )
         indx = state->code.fetchShort();
      else
         indx = state->code.fetchByte();
      state->locals[indx] = state->pop();
      break;
   }

   case ASTORE_0: case ASTORE_1: case ASTORE_2: case ASTORE_3:
      state->locals[oc-ASTORE_0] = state->pop();
      break;

   case ASTORE: {
      JavaUInt indx;
      if( wideFlag )
         indx = state->code.fetchShort();
      else
         indx = state->code.fetchByte();
      state->locals[indx] = state->pop();
      break;
   }

   case WIDE:
      wideFlag = true;
      goto AGAIN;

   case DUP: {
      JavaInt val = state->pop();
      state->push(val);
      state->push(val);
      break;
   }

   case DUP2: {
      JavaInt val1 = state->pop();
      JavaInt val2 = state->pop();
      state->push(val2);
      state->push(val1);
      state->push(val2);
      state->push(val1);
      break;
   }

   case POP:
      state->pop();
      break;

//   case NEW: case ANEWARRAY:
   case NEWARRAY: {
      JavaUByte indx = state->code.fetchByte();
      procNewArray(indx);
      break;
   }

   case IALOAD: {
      long index = state->pop();
      CJavaArray* ap = (CJavaArray*)state->pop();
      state->push(ap->get(index));
      break;
   }

   case IASTORE: {
      long value = state->pop();
      long index = state->pop();
      CJavaArray* ap = (CJavaArray*)state->pop();
      ap->put(index, value);
      break;      
   }

   case RETURN: case IRETURN:
      procReturnOps(oc);
      break;

   default:
      fprintf(stderr, "Chyba: nespravny kod instrukce - %u\n", oc);
      break;
   }
} // JavaVM::step

